/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/recipies/Recipies.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Convert.hpp"

void Aidge::explicitConvert(std::shared_ptr<GraphView> graph) {
    const auto nodes = graph->getNodes();
    for (auto node : nodes) {
        // TODO: currently, Operator data type is only reflected in its output tensor data type.
        // But an Operator might have multiple outputs of different data type(?)
        const auto& output = std::static_pointer_cast<OperatorTensor>(node->getOperator())->getOutput(0);
        if (output->getImpl() == nullptr) {
            continue;
        }
        const auto& device = output->getImpl()->device();

        if (node->type() == Convert_Op::Type) {
            // Remove existing Convert operator, if not needed anymore
            const auto parent = node->inputs()[0];
            const auto& input = std::static_pointer_cast<OperatorTensor>(parent.first->getOperator())->getOutput(parent.second);

            if (input->dataType() == output->dataType()
                && (input->getImpl()->device() == device))
            {
                // Add direct connection bypassing Convert node
                const auto childs = node->outputs()[0];
                for (const auto& child : childs) {
                    parent.first->addChild(child.first, parent.second, child.second);
                }

                // Remove all Convert node connections
                node->resetConnections();
            }
        }
        else {
            // Insert Convert operator between node inputs and parent output, if needed
            IOIndex_t inputIdx = 0;
            for (auto parent : node->inputs()) {
                if (parent.first != nullptr) {
                    const auto& input = std::static_pointer_cast<OperatorTensor>(parent.first->getOperator())->getOutput(parent.second);
        
                    if (input->dataType() != output->dataType()
                        || (input->getImpl()->device() != device))
                    {
                        // A conversion Operator is needed
                        auto convert = Convert();
                        convert->addChild(node, 0, inputIdx);
                        parent.first->addChild(convert, parent.second, 0);
                        // Set backend AFTER connection in case a specific implementation
                        // of the operator exists for the input type.
                        convert->getOperator()->setBackend(device.first, device.second);

                        // Add/update nodes in the GraphView
                        graph->add(convert);
                        graph->add(parent.first);
                        graph->add(node);
                    }
                }

                ++inputIdx;
            }
        }
    }
}
